package org.highpoint.titus.apps.osp.canonicalspaceship;

import java.awt.*;
import org.opensourcephysics.display.*;
import org.opensourcephysics.numerics.*;

/**
 * CanonicalSpaceShip implements the ODE solver to determine the position
 * and velocity of the spaceship.
 *
 *
 * This is an OSP application (www.opensourcephysics.org). It is governed by
 * the same license as OSP which is the GNU General Public License. A copy of
 * the GPL license is distributed with this source in
 * org/highpoint/titus/respanources/gpl.txt.

 * Copyright (C) 2004  Aaron Titus

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *
 */

/**
 * @author Aaron Titus
 *
 */


public class CanonicalSpaceShip
    implements ODE {
  private ODESolver ode_method;

  int pixRadius = 6;
  Color color = Color.red;
  double[] state;

  /**
   * The  constructor.
   */

  public CanonicalSpaceShip() {
    state = new double[12];
    ode_method = new RK45MultiStep(this);
  }

  private double force(double r) {
    if (r == 0) {
      return 0;
    }
    return 1 / r / r;
  }

  /**
   * Get the state of the spaceship.
   *
   * The state[] array contains: position, velocity, time, and thrust
   *
   * @return  the state
   */
  public double[] getState() {
    return state;
  }

  public void getRate(double[] state, double[] rate) {
    rate[0] = state[3]; // the rate of change x
    rate[1] = state[4]; // the rate of change y
    rate[2] = state[5]; // the rate of change z
    if (state[7] == 1 && state[8] == 0) {
      rate[4] = -state[11]; // vy rate is in -y direction
    }
    else if (state[7] == 0 && state[8] == 1) {
      rate[4] = state[11]; // vy rate is in +y direction
    }
    else {
      rate[4] = 0;
    } // vy rate is 0 +y direction

    if (state[9] == 1 && state[10] == 0) {
      rate[3] = -state[11]; // vx rate is in -x direction
    }
    else if (state[9] == 0 && state[10] == 1) {
      rate[3] = state[11]; // vx rate is in +x direction
    }
    else {
      rate[3] = 0;
    } // vy rate is 0
    rate[5] = 0; // vz rate
    rate[6] = 1; // time rate
    rate[7] = 0;
    rate[8] = 0;
    rate[9] = 0;
    rate[10] = 0;
    rate[11] = 0;
  }

  /**
   * Initialize the spaceship
   *
   * @param drawingPanel the drawing panel that will dispaly the particles
   * @param x x position
   * @param y y position
   * @param vx x velocity
   * @param vy y velocity
   * @param vz z velocity
   * @param dt time step for ODE solver
   * @param thrustN on/off (1 or 0) for N engine
   * @param thrustS on/off (1 or 0) for S engine
   * @param thrustE on/off (1 or 0) for E engine
   * @param thrustW on/off (1 or 0) for W engine
   * @param magThrust magnitude of engine thrust
   * @return  the state variables
   */
  public void initialize(
      DrawingPanel drawingPanel,
      double x,
      double y,
      double z,
      double vx,
      double vy,
      double vz,
      double dt,
      boolean thrustN,
      boolean thrustS,
      boolean thrustE,
      boolean thrustW,
      double magThrust) {
    ode_method.setStepSize(dt);
    state[0] = x;
    state[1] = y;
    state[2] = z;
    state[3] = vx;
    state[4] = vy;
    state[5] = vz;
    state[6] = 0; // time
    if (thrustN == true) {
      state[7] = 1;
    }
    else {
      state[7] = 0;
    }
    if (thrustS == true) {
      state[8] = 1;
    }
    else {
      state[8] = 0;
    }
    if (thrustE == true) {
      state[9] = 1;
    }
    else {
      state[9] = 0;
    }
    if (thrustW == true) {
      state[10] = 1;
    }
    else {
      state[10] = 0;
    }
    state[11] = magThrust;
  }

  /**
   * Step the state forward in time.
   */
  public void stepState() {
    ode_method.step();
  }

}
