package org.highpoint.titus.apps.osp.canonicalspaceship;

import java.awt.Color;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import javax.swing.JFrame;

import org.opensourcephysics.controls.Animation;
import org.opensourcephysics.controls.AnimationControl;
import org.opensourcephysics.controls.Control;
import org.opensourcephysics.display.*;
import java.awt.image.BufferedImage;
import org.opensourcephysics.tools.ResourceLoader;
import org.opensourcephysics.display.Arrow;

/*
 * CanonicalSpaceShipApp allows students to experiment with force and motion
 * by firing engines of a space ship in the +x, +y, -x, or -y directions.
 *
 * This is an OSP application (www.opensourcephysics.org). It is governed by
 * the same license as OSP which is the GNU General Public License. A copy of
 * the GPL license is distributed with this source in
 * org/highpoint/titus/resources/gpl.txt.


 * Copyright (C) 2004  Aaron Titus

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *
 */


/**
 * @author Aaron Titus
 *
 */


public class CanonicalSpaceShipApp
    implements Animation, Runnable, KeyListener {

  Control myControl;
  Thread animationThread;
  DrawingPanel panel = new DrawingPanel();
  DrawingFrame frame = new DrawingFrame(panel);
  CanonicalSpaceShip system = new CanonicalSpaceShip();
  
  DrawableGroup group = new DrawableGroup();
  Arrow vVector = new Arrow(0,0,0,0);

  double dt = 0.05; //time increment
  double[] r = {
      0, 0, 0}; //initial position
  double[] v = {
      0, 0, 0}; //initial velocity
  boolean[] thrust = {
      false, false, false, false}; //thrusters on or off, 0 or 1
  double magThrust = 10;
  boolean wrap = true;
  boolean showVelocityVector=false;
  int screenWidthX = 30;
  int screenWidthY = 30;

  int probeSize = 8;

  String imagePath = "org/highpoint/titus/resources/images/spaceship/";
  String image=imagePath + "ship0.gif";
  BufferedImage probeImg = ResourceLoader.getBufferedImage(image);
  MeasuredImage probe = new MeasuredImage(probeImg,-probeSize, probeSize, -probeSize,probeSize);
  
  /**
   * Stores a reference to this program's control.
   * @param control Control
   */
  public void setControl(Control control) {
    myControl = control;
    resetAnimation();
  }

  /**
   * Set the default parameters in the control.
   */
  public void resetAnimation() {
    stopAnimation();
    myControl.clearMessages();
    myControl.setValue("x", r[0]);
    myControl.setValue("y", r[1]);
    myControl.setValue("z", r[2]);
    myControl.setValue("vx", v[0]);
    myControl.setValue("vy", v[1]);
    myControl.setValue("vz", v[2]);
    myControl.setValue("top thruster", thrust[0]);
    myControl.setValue("bottom thruster", thrust[1]);
    myControl.setValue("right thruster", thrust[2]);
    myControl.setValue("left thruster", thrust[3]);
    myControl.setValue("thrust magnitude", magThrust);
    myControl.setValue("dt", dt);
    myControl.setValue("wrap", wrap);
    myControl.setValue("show velocity vector", showVelocityVector);
    myControl.println("Click on the panel to give it focus. \nUse the following keys to fire the rocket engines:\n\n \t J - top engine \n \t K - fire bottom engine \n \t F - fire right engine \n \t D - fire left engine");
    system.initialize(panel, r[0], r[1], r[2], v[0], v[1], v[2], dt, thrust[0],
                      thrust[1], thrust[2], thrust[3], magThrust);
    panel.repaint();
  }

  /**
   * Read parameters from the control start the calculation.
   */
  public void initializeAnimation() {
    r[0] = myControl.getDouble("x");
    r[1] = myControl.getDouble("y");
    r[2] = myControl.getDouble("z");
    v[0] = myControl.getDouble("vx");
    v[1] = myControl.getDouble("vy");
    v[2] = myControl.getDouble("vz");
    thrust[0] = myControl.getBoolean("top thruster");
    thrust[1] = myControl.getBoolean("bottom thruster");
    thrust[2] = myControl.getBoolean("right thruster");
    thrust[3] = myControl.getBoolean("left thruster");
    magThrust = myControl.getDouble("thrust magnitude");
    wrap = myControl.getBoolean("wrap");
    showVelocityVector = myControl.getBoolean("show velocity vector");
    dt = myControl.getDouble("dt");
    system.initialize(panel, r[0], r[1], r[2], v[0], v[1], v[2], dt, thrust[0],
                      thrust[1], thrust[2], thrust[3], magThrust);
    group.setXY(r[0], r[1]);
    if(showVelocityVector==true) {
      vVector.setXlength(v[0]);
      vVector.setYlength(v[1]);
    }
    panel.repaint();
  }

  /**
   * Read parameters from the control and start the calculation.
   */
  public void startAnimation() {
    if (animationThread != null) {
      return; // the animation is running
    }
    animationThread = new Thread(this);
    animationThread.setDaemon(true);
    animationThread.start(); // start the animation
  }

  /**
   * Stop the animation.
   */
  public void stopAnimation() {
    Thread runningThread = animationThread;
    animationThread = null; // stop the animation
    try { // wait for the thread to die
      if (runningThread != null) {
        runningThread.interrupt();
        runningThread.join();
      }
    }
    catch (InterruptedException e) {
    }
    ;
  }

  /**
   * Stop the animation.
   */

  public void stepAnimation() {
    if (animationThread != null) { // the animation is running so stop it
      stopAnimation();
      return;
    }
    system.stepState();
    if ( (myControl.getBoolean("wrap") == true) && (isOffScreen() == true)) {
      double x = system.state[0];
      double y = system.state[1];

      double xMax = screenWidthX;
      double yMax = screenWidthY;

      if (x > xMax || x < -xMax) {
        system.state[0] = -x;
      }
      else if (y > yMax || y < -yMax) {
        system.state[1] = -y;
      }

    }
    group.setXY(system.state[0], system.state[1]);
    panel.repaint();
  }

  /**
   * The run method is invoked by the animation thread.  DO NOT access this method.
   */
  public void run() {
    while (animationThread == Thread.currentThread()) {
      system.stepState();
      //check to see if the spaceship is off screen and if so, wrap it to the other side
      if ( (myControl.getBoolean("wrap") == true) && (isOffScreen() == true)) {
        double x = system.state[0];
        double y = system.state[1];

        double xMax = screenWidthX;
        double yMax = screenWidthY;

        if (x > xMax || x < -xMax) {
          system.state[0] = -x;
        }
        else if (y > yMax || y < -yMax) {
          system.state[1] = -y;
        }

      }

      //change the image of the ship depending on which engines are firing
      //note that these images will not load from within a jar file
      if (system.state[7] == 0 && system.state[8] == 0 && system.state[9] == 1 &&
          system.state[10] == 0) {
        setImage(imagePath + "ship1.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 1 &&
               system.state[9] == 0 && system.state[10] == 0) {
        setImage(imagePath + "ship2.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 1 &&
               system.state[9] == 1 && system.state[10] == 0) {
        setImage(imagePath + "ship3.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 0 &&
               system.state[9] == 0 && system.state[10] == 1) {
        setImage(imagePath + "ship4.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 0 &&
               system.state[9] == 1 && system.state[10] == 1) {
        setImage(imagePath + "ship5.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 1 &&
               system.state[9] == 0 && system.state[10] == 1) {
        setImage(imagePath + "ship6.gif");
      }
      else if (system.state[7] == 0 && system.state[8] == 1 &&
               system.state[9] == 1 && system.state[10] == 1) {
        setImage(imagePath + "ship7.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 0 &&
               system.state[9] == 0 && system.state[10] == 0) {
        setImage(imagePath + "ship8.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 0 &&
               system.state[9] == 1 && system.state[10] == 0) {
        setImage(imagePath + "ship9.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 1 &&
               system.state[9] == 0 && system.state[10] == 0) {
        setImage(imagePath + "ship10.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 1 &&
               system.state[9] == 1 && system.state[10] == 0) {
        setImage(imagePath + "ship11.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 0 &&
               system.state[9] == 0 && system.state[10] == 1) {
        setImage(imagePath + "ship12.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 0 &&
               system.state[9] == 1 && system.state[10] == 1) {
        setImage(imagePath + "ship13.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 1 &&
               system.state[9] == 0 && system.state[10] == 1) {
        setImage(imagePath + "ship14.gif");
      }
      else if (system.state[7] == 1 && system.state[8] == 1 &&
               system.state[9] == 1 && system.state[10] == 1) {
        setImage(imagePath + "ship15.gif");
      }
      else {
        setImage(imagePath + "ship0.gif");
      }

      group.setXY(system.state[0], system.state[1]);
      panel.render();
      try {
        animationThread.sleep(100);
      }
      catch (InterruptedException ie) {
      }
    }
  }

  /**
   * Checks to see if the spaceship is off the screen.
   *
   */

  public boolean isOffScreen() {
    if (system.state[0] > (double) screenWidthX ||
        system.state[0] < (double) - screenWidthX ||
        system.state[1] > (double) screenWidthY ||
        system.state[1] < (double) - screenWidthY) {
      return true;
    }
    else {
      return false;
    }
  }

  /* (non-Javadoc)
   * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
   */
  public void keyPressed(KeyEvent e) {
    int key = e.getKeyCode();
    // keyboard code for the key that was pressed
    //thrustN is state[7]
    //thrustS is state[8]
    //thrustE is state[9]
    //thrustW is state[10]

    if (key == KeyEvent.VK_J) {
      system.state[8] = 1;
    }
    if (key == KeyEvent.VK_K) {
      system.state[7] = 1;
    }
    if (key == KeyEvent.VK_F) {
      system.state[9] = 1;
    }
    if (key == KeyEvent.VK_D) {
      system.state[10] = 1;
    }

  }

  /* (non-Javadoc)
   * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
   */
  public void keyReleased(KeyEvent e) {
    int key = e.getKeyCode();
    // keyboard code for the key that was pressed
    //thrustN is state[7]
    //thrustS is state[8]
    //thrustE is state[9]
    //thrustW is state[10]

    if (key == KeyEvent.VK_J) {
      system.state[8] = 0;
    }
    if (key == KeyEvent.VK_K) {
      system.state[7] = 0;
    }
    if (key == KeyEvent.VK_F) {
      system.state[9] = 0;
    }
    if (key == KeyEvent.VK_D) {
      system.state[10] = 0;
    }

  }

  /* (non-Javadoc)
   * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
   */
  public void keyTyped(KeyEvent e) {
  }

  /*
   * set the image of the spaceship
   * @param filename filename relative to the working directory
   */

  public void setImage(String filename) {
    probeImg = ResourceLoader.getBufferedImage(filename);
    probe.setImage(probeImg);
  }

  /**
   * Constructor
   *
   */

  public CanonicalSpaceShipApp() {
    panel.setPreferredMinMax( -1 * screenWidthX, screenWidthX,
                             -1 * screenWidthY, screenWidthY);
    panel.setSquareAspect(false);
    panel.addDrawable(group);
    group.addDrawable(probe);
    panel.setBackground(Color.black);
    frame.addKeyListener(this); //note that the key listener must be added to the frame, not the panel because the panel never has focus
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.setSize(screenWidthX * 20, screenWidthY * 20);
    panel.repaint();
    
    if(showVelocityVector==true) {
      group.addDrawable(vVector);
    }

  }

  /**
   * Starts the Java application.
   * @param args  command line parameters
   */

  public static void main(String[] args) {
    Animation app = new CanonicalSpaceShipApp();
    AnimationControl control = new AnimationControl(app);
    app.setControl(control);
  }

}
