/*
 * The org.opensourcephysics.display3d.simple3d package implements the
 * classes of the org.opensourcephysics.display3d package using
 * the so called painter's algorithm, in which the diferent parts of the
 * 3D scene are painted from back to front, thus achieving a simple, but
 * for many cases effective, hidden-line removal.
 *
 * Copyright (c) 2005  The Open Source Physics project
 *                     http://www.opensourcephysics.org
 */

package org.opensourcephysics.display3d.simple3d;

import java.awt.*;
import org.opensourcephysics.controls.*;
import java.awt.geom.AffineTransform;

/**
 * <p>Title: ElementCircle</p>
 * <p>Description: A Circle using the painter's algorithm</p>
 * @author Francisco Esquembre
 * @version March 2005
 */
public class ElementCircle extends Element implements org.opensourcephysics.display3d.core.ElementCircle {
  // Configuration variables
  private double angle = 0.0;

  // Implementation variables
  private double[] coordinates = new double[3];
  private double[] size = new double[3];
  private double[] pixel  = new double[3]; // The ouput of position projections
  private double[] pixelSize  = new double[2]; // The ouput of size projections
  private Object3D[] objects = new Object3D[] { new Object3D(this,0) };
  private AffineTransform transform = new AffineTransform(), originalTransform=null;
  {
    setSizeXYZ(0,0,0);
  }
// -------------------------------------
// New configuration methods
// -------------------------------------

  public void setRotationAngle (double angle) { this.angle = angle; }
  public double getRotationAngle () { return this.angle; }

// -------------------------------------
// Abstract part of Element
// -------------------------------------

  Object3D[] getObjects3D() {
    if (!isVisible()) return null;
    if (hasChanged() || needsToProject()) projectPoints ();
    return objects;
  }

  void draw (Graphics2D _g2, int _index) {
    // Allow the panel to adjust color according to depth
    Color theColor = getPanel().projectColor(getRealStyle().getLineColor(),objects[0].getDistance());
    Color theFillColor = getPanel().projectColor(getRealStyle().getFillColor(),objects[0].getDistance());
    drawIt (_g2,theColor,theFillColor);
  }

  void drawQuickly (Graphics2D _g2) {
    if (!isVisible()) return;
    if (hasChanged() || needsToProject()) projectPoints ();
    drawIt (_g2, getRealStyle().getLineColor(),null); // getRealStyle().getFillColor());
  }

  void styleChanged (int styleThatChanged) { } // Not affected

// -------------------------------------
// Interaction
// -------------------------------------

  protected InteractionTarget getTargetHit (int x, int y) {
    if (!isVisible()) return null;
    if (hasChanged() || needsToProject()) projectPoints ();
    if (targetPosition.isEnabled() && Math.abs(pixel[0]-x)<SENSIBILITY && Math.abs(pixel[1]-y)<SENSIBILITY) return targetPosition;
    return null;
   }

// -------------------------------------
// Private methods
// -------------------------------------

  private void projectPoints () {
    coordinates[0] = coordinates[1] = coordinates[2] = 0.0;
    sizeAndToSpaceFrame (coordinates);
    getPanel().project(coordinates,pixel);
    objects[0].setDistance(pixel[2]);
    size[0] = getSizeX(); size[1] = getSizeY(); size[2] = getSizeZ();
    getPanel().projectSize(coordinates,size,pixelSize);
    setElementChanged(false);
    setNeedToProject(false);
  }

  private void drawIt (Graphics2D _g2, Color _color, Color _fill) {
    int xc = (int)(pixel[0]-pixelSize[0]/2), yc = (int)(pixel[1]-pixelSize[1]/2);
    _g2.setStroke(getRealStyle().getLineStroke());
    if (angle!=0.0) {
      originalTransform = _g2.getTransform();
      transform.setTransform(originalTransform);
      transform.rotate(-angle, pixel[0], pixel[1]);
      _g2.setTransform(transform);
    }
    if (_fill!=null) { // First fill the inside
      _g2.setPaint(_fill);
      _g2.fillOval(xc,yc,(int)pixelSize[0]+1,(int)pixelSize[1]+1); // Is this a bug in Awt?
    }
    _g2.setColor (_color);
    _g2.drawOval(xc,yc,(int)pixelSize[0],(int)pixelSize[1]);
    if (angle!=0.0) _g2.setTransform(originalTransform);
  }

// ----------------------------------------------------
// XML loader
// ----------------------------------------------------

  /**
   * Returns an XML.ObjectLoader to save and load object data.
   * @return the XML.ObjectLoader
   */
  public static XML.ObjectLoader getLoader() { return new ElementCircleLoader(); }

  static private class ElementCircleLoader extends ElementLoader {

    public Object createObject(XMLControl control) { return new ElementCircle(); }

    public void saveObject(XMLControl control, Object obj) {
      super.saveObject(control,obj);
      ElementImage element = (ElementImage) obj;
      control.setValue("rotation angle", element.getRotationAngle());
   }

    public Object loadObject(XMLControl control, Object obj) {
      super.loadObject(control,obj);
      ElementImage element = (ElementImage) obj;
      element.setRotationAngle(control.getDouble("rotation angle"));
      return obj;
    }

  }

}
