package org.opensourcephysics.numerics;
import org.opensourcephysics.controls.*;

/**
 * Matrix3DTransformation implements 3D affine transformations using a matrix representation.
 */
public class Matrix3DTransformation implements Transformation {

   private double[][] matrix = new double[3][3]; // the transformation matrix
   private double[][] inverse = null;            // the inverse transformation matrix if it exists

   /**
    * Constructs a 3D transformation using the given matrix.
    *
    * Affine transformations can be applied to 3D coordinates.
    * A 3 by 3 matrix sets the rotation and shear.
    * A null matrix sets the transformation to the identity transformation.
    *
    * @param mat double[][]
    */

   public Matrix3DTransformation(double[][] matrix) {
      if(matrix==null) { // identiy matrix
         this.matrix[0][0] = this.matrix[1][1] = this.matrix[2][2] = 1;
         return;
      }
      for(int i = 0; i<matrix.length; i++) { // loop over the rows
         System.arraycopy(matrix[i], 0, this.matrix[i], 0, matrix[i].length);
      }
   }

   /**
    * Creates a 3D transforamtion representing a rotation about the x axis by the given angle.
    *
    * @param theta double
    *
    * @return Affine3DTransformation
    */
   public static Matrix3DTransformation RotationX(double theta){
      //same as Rotation(theta, new double[]{0, 0, 1});
      Matrix3DTransformation at= new Matrix3DTransformation(null); // creates matrix
      double[][] mat = at.matrix;
      double sin=Math.sin(theta);
      double cos=Math.cos(theta);
      // matrix elements not listed are zero
      mat[0][0] = 1;
      mat[1][1] = cos;
      mat[1][2] = -sin;
      mat[2][1] = sin;
      mat[2][2] = cos;
      // inverse matrix is null but we know what it is so create it
      double[][] inv = new double[3][3];
      at.inverse=inv;
      inv[0][0] = 1;
      inv[1][1] = cos;
      inv[1][2] = sin;
      inv[2][1] = -sin;
      inv[2][2] = cos;
      return at;
   }

   /**
    * Creates a 3D transforamtion representing a rotation about the y axis by the given angle.
    *
    * @param theta double
    *
    * @return Affine3DTransformation
    */
   public static Matrix3DTransformation RotationY(double theta){
      //same as Rotation(theta, new double[]{0, 0, 1});
      Matrix3DTransformation at= new Matrix3DTransformation(null); // creates matrix
      double[][] mat = at.matrix;
      double sin=Math.sin(theta);
      double cos=Math.cos(theta);
      // matrix elements not listed are zero
      mat[1][1] = 1;
      mat[0][0] = cos;
      mat[0][2] = sin;
      mat[2][0] = -sin;
      mat[2][2] = cos;
      // inverse matrix is null but we know what it is so create it
      double[][] inv = new double[3][3];
      at.inverse=inv;
      // matrix elements not listed are zero
      inv[1][1] = 1;
      inv[0][0] = cos;
      inv[0][2] = -sin;
      inv[2][0] = sin;
      inv[2][2] = cos;
      return at;
   }

   /**
    * Creates a 3D transforamtion representing a rotation about the z axis by the given angle.
    *
    * @param theta double
    *
    * @return Affine3DTransformation
    */
   public static Matrix3DTransformation RotationZ(double theta){
      //same as Rotation(theta, new double[]{0, 0, 1});
      Matrix3DTransformation at= new Matrix3DTransformation(null); // creates matrix
      double[][] mat = at.matrix;
      double sin=Math.sin(theta);
      double cos=Math.cos(theta);
      // matrix elements not listed are zero
      mat[0][0] = cos;
      mat[0][1] = -sin;
      mat[1][0] = sin;
      mat[1][1] = cos;
      mat[2][2] = 1;
      // inverse matrix is null but we know what it is so create it
      double[][] inv = new double[3][3];
      at.inverse=inv;
      inv[0][0] = cos;
      inv[0][1] = sin;
      inv[1][0] = -sin;
      inv[1][1] = cos;
      inv[2][2] = 1;
      return at;
   }


   /**
    * Creates a 3D transforamtion representing a rotation about the origin by the given angle around
    * the given axis.
    *
    * @param theta double
    * @param axis double[]
    * @return Affine3DTransformation
    */
   public static Matrix3DTransformation Rotation(double theta, double[] axis) {
      Matrix3DTransformation at = new Matrix3DTransformation(null); // creates unit matrix
      double[][] mat = at.matrix;
      double x = axis[0], y = axis[1], z = axis[2];
      double norm = x*x+y*y+z*z;
       if(norm!=1){ // this usually doesn't happen because of roundoff but is worth a try
         norm=1/Math.sqrt(norm);
         x *= norm;
         y *= norm;
         z *= norm;
      }
      double c = Math.cos(theta), s = Math.sin(theta);
      double t = 1-c;
      // matrix elements not listed are zero
      mat[0][0] = t*x*x+c;
      mat[0][1] = t*x*y-s*z;
      mat[0][2] = t*x*z+s*y;
      mat[1][0] = t*x*y+s*z;
      mat[1][1] = t*y*y+c;
      mat[1][2] = t*y*z-s*x;
      mat[2][0] = t*x*z-s*y;
      mat[2][1] = t*y*z+s*x;
      mat[2][2] = t*z*z+c;
      // inverse matrix is null but we know what it is so create it
      double[][] inv = new double[3][3];
      at.inverse = inv;
      inv[0][0] = mat[0][0];
      inv[1][0] = mat[0][1];
      inv[2][0] = mat[0][2];
      inv[0][1] = mat[1][0];
      inv[1][1] = mat[1][1];
      inv[2][1] = mat[1][2];
      inv[0][2] = mat[2][0];
      inv[1][2] = mat[2][1];
      inv[2][2] = mat[2][2];
      return at;
   }

   /**
 * Creates an transformation representing a rotation about the origin by the given quaternion.
 *
 * @param quaternion double[]
 * @return Affine3DTransformation
 */
public static Matrix3DTransformation Quaternion(double[] quaternion) {
   return Quaternion(quaternion[0],quaternion[1],quaternion[2],quaternion[3]);
}

/**
 * Creates an AffineMatrix representing a rotation about the origin by the given quaternion components.
 *
 * @param quaternion double[]
 * @return Affine3DTransformation
 */
public static Matrix3DTransformation Quaternion(double q0, double q1, double q2, double q3){
   Matrix3DTransformation at = new Matrix3DTransformation(null);
   double[][] atMatrix = at.matrix;
   double norm = q0*q0+q1*q1+q2*q2+q3*q3;
   if(norm!=1){ // this usually doesn't happen because of roundoff but is worth a try
      norm = 1/Math.sqrt(norm); // computationaly expensive
      q0 *= norm;
      q1 *= norm;
      q2 *= norm;
      q3 *= norm;
   }
   double q11 = 2*q1*q1, q22 = 2*q2*q2, q33 = 2*q3*q3;
   double q12 = 2*q1*q2, q13 = 2*q1*q3, q23 = 2*q2*q3;
   double q01 = 2*q0*q1, q02 = 2*q0*q2, q03 = 2*q0*q3;

// matrix elements not listed are zero
   atMatrix[0][0] = 1-q22-q33;
   atMatrix[0][1] = q12-q03;
   atMatrix[0][2] = q13+q02;
   atMatrix[1][0] = q12+q03;
   atMatrix[1][1] = 1-q11-q33;
   atMatrix[1][2] = q23-q01;
   atMatrix[2][0] = q13-q02;
   atMatrix[2][1] = q23+q01;
   atMatrix[2][2] = 1-q11-q22;

   // inverse matrix is null but we know what it is so create it
   double[][] inv = new double[3][3];
   at.inverse = inv;
   inv[0][0] = atMatrix[0][0];
   inv[1][0] = atMatrix[0][1];
   inv[2][0] = atMatrix[0][2];
   inv[0][1] = atMatrix[1][0];
   inv[1][1] = atMatrix[1][1];
   inv[2][1] = atMatrix[1][2];
   inv[0][2] = atMatrix[2][0];
   inv[1][2] = atMatrix[2][1];
   inv[2][2] = atMatrix[2][2];
   return at;
}

   /**
    * Provides a copy of this transformation.
    */
   public Object clone() {
      return new Matrix3DTransformation(matrix);
   }

   /**
    * Gets the rotation that aligns the first vector with the second vector.
    *
    * @param v1 double[]
    * @param v2 double[]
    * @return Quaternion
    */
   public static Matrix3DTransformation getAlignment(double[] v1, double v2[]){
      v1 = VectorMath.normalize((double[]) v1.clone());
      v2 = VectorMath.normalize((double[]) v2.clone());
      double theta = Math.acos(VectorMath.dot(v1, v2));
      double[] axis = VectorMath.cross3D(v1, v2);
      return Matrix3DTransformation.Rotation(theta, axis);
   }


   /**
    * Transforms the given point.
    *
    * Convertes the point to homogeneous coordinates if the size of the point array is less than
    *
    * @param point the coordinates to be transformed
    */
   public double[] direct(double[] point) {
      double[] tempPoint= (double[])point.clone();
      for(int i = 0,n = point.length; i<n; i++) {
         point[i] = 0;
         for(int j = 0; j<n; j++) {
            point[i] += matrix[i][j]*tempPoint[j];
         }
      }
      return point;
   }

   /**
    * Transforms the given point using the inverse transformation (if it exists).
    *
    * If the transformation is not invertible, then a call to this
    * method must throw a UnsupportedOperationException exception.
    *
    * @param point the coordinates to be transformed
    */
   public double[] inverse(double[] point) throws UnsupportedOperationException {
      if (inverse==null){
         calcInverse(); // computes the inverse using LU decompostion
         if (inverse==null){ // inverse does not exist
            throw new UnsupportedOperationException("The inverse matrix does not exist.");
         }

      }
      double[] tempPoint= (double[])point.clone();
      for(int i = 0, n = point.length; i<n; i++) {
         point[i] = 0;
         for(int j = 0; j<n; j++) {
            point[i] += inverse[i][j]*tempPoint[j];
         }
      }
      return point;
   }

   private void calcInverse() {
      LUPDecomposition lupd = new LUPDecomposition(matrix);
      inverse = lupd.inverseMatrixComponents();
   }

   public static XML.ObjectLoader getLoader() {
      return new Affine3DTransformationLoader();
   }

   protected static class Affine3DTransformationLoader extends XMLLoader {

      public void saveObject(XMLControl control, Object obj) {
         Matrix3DTransformation transf = (Matrix3DTransformation) obj;
         control.setValue("matrix", transf.matrix);
         if(transf.inverse!=null)control.setValue("inverse", transf.inverse);
      }

      public Object createObject(XMLControl control) {
         return new Matrix3DTransformation(null);
      }

      public Object loadObject(XMLControl control, Object obj) {
         Matrix3DTransformation transf = (Matrix3DTransformation) obj;
         transf.matrix = (double[][]) control.getObject("matrix");
         transf.inverse = (double[][]) control.getObject("inverse");
         return obj;
      }
   }
}
