package org.opensourcephysics.numerics;

/**
 * Quaterion models a unit quaternion and implements quaternion arithmetic.
 */
public class Quaternion implements Transformation {
   static final double SQRT2=Math.sqrt(2);
   double q0, q1, q2, q3;

   /**
    * Constructs and initializes quaternion from the specified components.
    *
    * @param q0 double
    * @param q1 double
    * @param q2 double
    * @param q3 double
    */
   public Quaternion(double q0, double q1, double q2, double q3) {
      this.q0 = q0;
      this.q1 = q1;
      this.q2 = q2;
      this.q3 = q3;
      normalize();
   }

   /**
    * Constructs and initializes a quaternion from the array of length 4.
    * @param q the array of length 4 containing q0, q1, q2, q3
    */
   public Quaternion(double[] q) {
      q0 = q[0];
      q1 = q[1];
      q2 = q[2];
      q3 = q[3];
      normalize();
   }

   /**
    * Constructs and initializes a Quaternion with the same values as the given quaternion.
    * @param q the Vector3d containing the initialization x y z data
    */
   public Quaternion(Quaternion q) {
      q0 = q.q0;
      q1 = q.q1;
      q2 = q.q2;
      q3 = q.q3;
      normalize();
   }

   /**
    * Constructs and initializes a unit quaternion (1,0,0,0).
    */
   public Quaternion() {
      this(1, 0, 0, 0);
   }

   /**
    * Gets the quaternion that aligns the first vector with the second vector.
    *
    * @param v1 double[]
    * @param v2 double[]
    * @return Quaternion
    */
   public static Quaternion getAlignment(double[] v1, double v2[]){
      v1 = VectorMath.normalize((double[]) v1.clone());
      v2 = VectorMath.normalize((double[]) v2.clone());
      double[] r = VectorMath.cross3D(v1, v2);
      double s = Math.sqrt(2*(1+VectorMath.dot(v1, v2)));
      return new Quaternion(s/2, r[0]/s, r[1]/s, r[2]/s);
   }


   /**
    * Sets the quaternion coordinates.
    *
    * @param q0 double
    * @param q1 double
    * @param q2 double
    * @param q3 double
    */
   public final void setCoordinates(double q0, double q1, double q2, double q3){
      this.q0 = q0;
      this.q1 = q1;
      this.q2 = q2;
      this.q3 = q3;
      normalize();
   }

   /**
    * Sets the quaternion coordinates from the array of length 4.
    * @param q the array of length 4 containing q0, q1, q2, q3
    */
   public final void setCoordinates(double[] q){
      q0 = q[0];
      q1 = q[1];
      q2 = q[2];
      q3 = q[3];
      normalize();
   }


   /**
    * Normalizes this quaternion in place.
    */
   public final void normalize() {
      double norm = q0*q0+q1*q1+q2*q2+q3*q3;
      if(norm==1) return;  // often doesn't happen due to roundoff
      norm = 1.0/Math.sqrt(norm);   // expensive operation
      q0 *= norm;
      q1 *= norm;
      q2 *= norm;
      q3 *= norm;
   }

   /**
    * Conjugates this quaternion in place.
    */
   public final void conjugate(){
      q1 = -q1;
      q2 = -q2;
      q3 = -q3;
   }

   /**
    * Concatenates this quaternion with the given quaternion.
    */
   public final void concatenate(Quaternion q){
      double w= q0*q.q0 - q1*q.q1 - q2*q.q2 - q3*q.q3;
      double x= q3*q.q2 - q2*q.q3 + q1*q.q0 + q0*q.q1;
      double y= q1*q.q3 - q3*q.q1 + q2*q.q0 + q0*q.q2;
      double z= q2*q.q1 - q1*q.q2 + q3*q.q0 + q0*q.q3;;
      q0=w;
      q1=x;
      q2=y;
      q3=z;
      normalize();
   }

   /**
    * Returns the dot product of this quaternion and quaternion q.
    * @param q the other quaternion
    * @return the dot product of this and q
    */
   public final double dot(Quaternion q) {
      return(q0*q.q0+ q1*q.q1+q2*q.q2+q3*q.q3);
   }

   /**
    * Returns the squared magnitude of this vector.
    * @return the squared magnitude
    */
   public final double magnitudeSquared() {
      return(q0*q0+q1*q1+q2*q2+q3*q3);
   }

   /**
    * Returns the magnitude of this quaternion.
    * @return the magnitude
    */
   public final double magnitude() {
      return Math.sqrt(q0*q0+q1*q1+q2*q2+q3*q3);
   }

   /**
    *    Returns the angle in radians between this quaternion and the given
    *    quaternion.
    *
    *    @param q    the other quaternion
    *    @return   the angle in radians in the range [0,PI]
    */
   public final double angle(Quaternion q) {
      double norm1=Math.sqrt(q1*q1+q2*q2+q3*q3);  // normalization for vector part of this quaternion
      double norm2=Math.sqrt(q.q1*q.q1+q.q2*q.q2+q.q3*q.q3); // normazliation for vector part for given quaternion
      double w= Math.sqrt(1+(q1*q.q1+q2*q.q2+q3*q.q3)/norm1/norm2);
      return(2*Math.acos(w/SQRT2));
   }

   public Object clone(){
      return new Quaternion(q0,q1,q2,q3);
   }

   public double[] direct(double[] p){ // assumes quaternion is normalized
      double pMult=2*q0*q0-1;
      double vMult=2*(q1*p[0]+q2*p[1]+q3*p[2]);
      double crossMult=2*q0;
      double x=pMult*p[0]+vMult*q1+crossMult*(q2*p[2]-q3*p[1]);
      double y=pMult*p[1]+vMult*q2+crossMult*(q3*p[0]-q1*p[2]);
      p[2]    =pMult*p[2]+vMult*q3+crossMult*(q1*p[1]-q2*p[0]);
      p[0] = x;
      p[1] = y;
      return p;
   }

   public double[] inverse(double[] p) throws UnsupportedOperationException{ // assumes quaternion is normalized
      double pMult=2*q0*q0-1;
      double vMult = 2*(q1*p[0]+q2*p[1]+q3*p[2]);
      double crossMult = -2*q0;  // inverse transformation changes sign of cross term
      double x = pMult*p[0]+vMult*q1+crossMult*(q2*p[2]-q3*p[1]);
      double y = pMult*p[1]+vMult*q2+crossMult*(q3*p[0]-q1*p[2]);
      p[2]     = pMult*p[2]+vMult*q3+crossMult*(q1*p[1]-q2*p[0]);
      p[0] = x;
      p[1] = y;
      return p;
   }

}
