/*
 * The org.opensourcephysics.numerics package contains numerical methods
 * for the book Simulations in Physics.
 * Copyright (c) 2005  H. Gould, J. Tobochnik, and W. Christian.
 */
package org.opensourcephysics.numerics;

/**
 * A utility class for numerical analysis.
 * This class cannot be subclassed or instantiated because all methods are static.
 *
 * @author       Wolfgang Christian
 * @version 1.0
 */
public final class Util {
   /** The default precision for numerical analysis. */
   public static final double defaultNumericalPrecision = Math.sqrt(Double.MIN_VALUE);
   /** Parser for simple arithmetic expressions. */
   private static SuryonoParser parser = new SuryonoParser(0); // parser without variables

   private Util() {} // prohibit instantiation because all methods are static

   /**
    * Computes the relativePrecision except near zero where the absolute precision is returned.
    *
    * @param epsilon the absolute error
    * @param result  the result
    *
    * @return the relative error
    */
   public static double relativePrecision(final double epsilon, final double result) {
      return(result>defaultNumericalPrecision)
            ? epsilon/result
            : epsilon;
   }

   /**
    * Checks if an array is sorted.
    * Returns:
    *   Positive integer if array is sorted in increasing value.
    *   Negative integer if array is sorted in decreasing value.
    *   Zero if array is not sorted.
    *
    * @param array double[]
    * @return int 1,0,-1 based on sorting
    */
   public static int checkSorting(double[] array) {
      int sign = (array[0]<=array[array.length-1])
                 ? 1
                 : -1;
      for(int i = 1, n = array.length; i<n; i++) {
         switch(sign) {
            case -1 :
               if(array[i-1]<array[i]) {
                  return 0; // not sorted
               }
               break;
            case 1 :
               if(array[i-1]>array[i]) {
                  return 0; // not sorted
               }
         }
      }
      return sign;
   }

   /**
    * Gets the approximate range of a function within the given domain.
    *
    * The range is deterermiend by evaluating the function at n points and finding the minimum and maximum values.
    *
    * @param f Function
    * @param a double
    * @param b double
    * @param n int
    * @return double[]
    */
   public static double[] getRange(Function f, double a, double b, int n) {
      double min = f.evaluate(a);
      double max = f.evaluate(a);
      double x = a, dx = (b-a)/(n-1);
      for(int i = 1; i<n; i++) {
         double y = f.evaluate(x);
         min = Math.min(min, y);
         max = Math.max(max, y);
         x += dx;
      }
      return new double[]{min, max};
   }

   /**
    * Fills the given double[2][n] array with x and f(x) values on interval [start, stop].
    *
    * @param f Function
    * @param start double
    * @param stop double
    * @param data double[][]
    * @return double[][]
    */
   static public double[][] functionFill(Function f, double start, double stop, double[][] data) {
      double dx = 1;
      int n = data[0].length;
      if(n>1) {
         dx = (stop-start)/(n-1);
      }
      double x = start;
      for(int i = 0; i<n; i++) {
         data[0][i] = x;
         data[1][i] = f.evaluate(x);
         x += dx;
      }
      return data;
   }

   /**
    * Fills the given double[n] array with f(x) on the interval [start, stop].
    *
    * @param f Function
    * @param start double starting value of x
    * @param stop double  stopping value of x
    * @param data double[]
    * @return double[]
    */
   static public double[] functionFill(Function f, double start, double stop, double[] data) {
      double dx = 1;
      int n = data.length;
      if(n>1) {
         dx = (stop-start)/(n-1);
      }
      double x = start;
      for(int i = 0; i<n; i++) {
         data[i] = f.evaluate(x);
         x += dx;
      }
      return data;
   }

   /**
    * Computes the average value of a subset of an array.
    *
    * @param array the data to be averaged
    * @param start the index of the first point to be averaged
    * @param num   the total number of points to be averaged
    * @return
    */
   public static double computeAverage(double[] array, int start, int num) {
      double sum = 0;
      for(int i = start, stop = start+num; i<stop; i++) {
         sum += array[i];
      }
      return sum/num;
   }

   /**
    * Creates a function having a constant value.
    * @param _c
    * @return
    */
   public static Function constantFunction(double _c) {
      final double c = _c;
      return new Function() {

         public double evaluate(double x) {
            return c;
         }
      };
   }

   /**
    * Evalautes a mathematical expression without variables.
    * @param str String
    * @return double
    */
   public static synchronized double evalMath(String str) {
      try {
         parser.parse(str);
         return parser.evaluate();
      } catch(ParserException ex) {}
      return Double.NaN;
   }
}
