/*
 * The org.opensourcephysics.tools package defines classes for managing OSP
 * applications and objects.
 */
package org.opensourcephysics.tools;

import java.io.*;
import java.util.*;

import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import org.opensourcephysics.controls.*;
import java.awt.event.*;

/**
 * This is a dialog that displays a tree showing tabs to be saved.
 *
 * @author Douglas Brown
 * @version 1.0
 */
public class LaunchSaver extends JDialog {

  // instance fields
  private String prevTabSetName;
  private String prevTabSetBasePath;
  private LaunchBuilder launcher;
  private JTree tree;
  private JTextField pathField;
  private JButton setButton;
  private SaverNode root;
  private JScrollPane treeScroller;
  private boolean approved = false;
  FocusListener focusListener;
  KeyListener keyListener;
  SaverEditor editor = new SaverEditor();

  /**
   * Constructor.
   *
   * @param launcher the Launcher with tabs to save
   */
  public LaunchSaver(LaunchBuilder launcher) {
    super(launcher.frame, true);
    createGUI();
    setLauncher(launcher);
    pack();
    // center on screen
    Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
    int x = (dim.width - getBounds().width) / 2;
    int y = (dim.height - getBounds().height) / 2;
    setLocation(x, y);
  }

  /**
   * Sets the Launcher.
   *
   * @param launcher the Launcher
   */
  public void setLauncher(LaunchBuilder launcher) {
    this.launcher = launcher;
    createTree();
    refresh();
    approved = false;
  }

  /**
   * Sets the selected node.
   *
   * @param node the node to select
   */
  public void setSelectedNode(SaverNode node) {
    if (node == null) return;
    tree.setSelectionPath(new TreePath(node.getPath()));
  }

  /**
   * Sets visiblity. Overrides superclass method.
   *
   * @param vis true to show
   */
  public void setVisible(boolean vis) {
    if (vis) {
      prevTabSetName = launcher.tabSetName;
      prevTabSetBasePath = launcher.tabSetBasePath;
    }
    super.setVisible(vis);
  }

  /**
   * Gets the selected node.
   *
   * @return the selected node
   */
  public SaverNode getSelectedNode() {
    TreePath path = tree.getSelectionPath();
    if (path == null) return null;
    else return (SaverNode)path.getLastPathComponent();
  }

  /**
   * Returns true if user has clicked the save button, otherwise false.
   *
   * @return true to save
   */
  public boolean isApproved() {
    return approved;
  }

  /**
   * Creates any missing folders in the specified path.
   *
   * @param path the path to construct
   */
  public void createFolders(String path) {
    // work way up path to find existing folder
    File dir = new File(path);
    ArrayList dirs = new ArrayList(); // list of needed directories
    while (!dir.exists()) {
      dirs.add(0, dir);
      int j = path.lastIndexOf("/");
      if (j == -1)break;
      path = path.substring(0, j);
      dir = new File(path);
    }
    // work back down path and create needed directories
    Iterator it = dirs.iterator();
    while (it.hasNext()) {
      dir = (File) it.next();
      dir.mkdir();
    }
  }

//______________________________ protected methods _____________________________

  /**
   * Refreshes the GUI.
   */
  private void refresh() {
    pathField.setText(
      XML.getResolvedPath(launcher.tabSetName, launcher.tabSetBasePath));
    pathField.setBackground(Color.white);
    tree.repaint();
  }

  /**
   * Reverts to initial state.
   */
  private void revert() {
    launcher.tabSetBasePath = prevTabSetBasePath;
    launcher.tabSetName = prevTabSetName;
  }

  /**
   * Creates the GUI.
   */
  protected void createGUI() {
      // create focus listener
    focusListener = new FocusAdapter() {
      public void focusLost(FocusEvent e) {
        if (getSelectedNode() == root) {
          String path = Launcher.forwardSlash(pathField.getText());
          if (!"xset".equals(Launcher.getExtension(path))) {
            while (path.indexOf(".") > -1) {
              path = path.substring(0, path.length() - 1);
            }
            path += ".xset";
          }
          launcher.tabSetBasePath = XML.getDirectoryPath(path);
          launcher.tabSetName = Launcher.getName(path);
        }
        refresh();
      }
    };
    // create key listener
    keyListener = new KeyAdapter() {
      public void keyPressed(KeyEvent e) {
        if (getSelectedNode() == root && e.getKeyCode() == KeyEvent.VK_ENTER) {
          String path = Launcher.forwardSlash(pathField.getText());
          if (!"xset".equals(Launcher.getExtension(path))) {
            while (path.indexOf(".") > -1) {
              path = path.substring(0, path.length() - 1);
            }
            path += ".xset";
          }
          launcher.tabSetBasePath = XML.getDirectoryPath(path);
          launcher.tabSetName = Launcher.getName(path);
          refresh();
        }
        else {
          pathField.setBackground(launcher.YELLOW);
        }
      }
    };
    setTitle(LaunchRes.getString("Saver.Title"));
    JPanel panel = new JPanel(new BorderLayout());
    panel.setPreferredSize(new Dimension(400, 200));
    setContentPane(panel);
    JPanel upper = new JPanel(new BorderLayout());
    JLabel text1 = new JLabel(LaunchRes.getString("Saver.Text.Line1"));
    JLabel text2 = new JLabel(LaunchRes.getString("Saver.Text.Line2"));
    text1.setBorder(BorderFactory.createEmptyBorder(4, 4, 1, 4));
    text2.setBorder(BorderFactory.createEmptyBorder(1, 4, 4, 4));
    upper.add(text1, BorderLayout.NORTH);
    upper.add(text2, BorderLayout.SOUTH);
    panel.add(upper, BorderLayout.NORTH);
    JPanel lower = new JPanel(new BorderLayout());
    panel.add(lower, BorderLayout.SOUTH);
    JToolBar toolbar = new JToolBar();
    toolbar.setFloatable(false);
    lower.add(toolbar, BorderLayout.NORTH);
    JLabel label = new JLabel(LaunchRes.getString("Saver.Label"));
    label.setBorder(BorderFactory.createEmptyBorder(0, 2, 0, 4));
    toolbar.add(label);
    pathField = new JTextField();
    pathField.addKeyListener(keyListener);
    pathField.addFocusListener(focusListener);
    toolbar.add(pathField);
    setButton = new JButton(LaunchRes.getString("Saver.Button.Set"));
    setButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        editor.getCellEditorValue(); // make sure edits are saved
        choosePath();
      }
    });
    JToolBar buttonbar = new JToolBar();
    buttonbar.setFloatable(false);
    buttonbar.add(Box.createHorizontalGlue());
    buttonbar.add(setButton);
    lower.add(buttonbar, BorderLayout.SOUTH);
    // save button
    JButton button = new JButton(LaunchRes.getString("Saver.Button.Save"));
    buttonbar.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        editor.getCellEditorValue(); // make sure edits are saved
        // set selfContained properties based on current tree state
        int n = tree.getRowCount();
        for (int i = 0; i < n; i++) {
          TreePath path = tree.getPathForRow(i);
          SaverNode node = (SaverNode)path.getLastPathComponent();
          if (tree.isCollapsed(path)) {
            if (node.getChildCount() > 0) { // this node is self contained
              if (node == root) {
                launcher.selfContained = true;
              }
              else {
                node.node.setSelfContained(true);
              }
            }
          }
          else {
            if (node == root) {
              launcher.selfContained = false;
            }
            else {
              node.node.setSelfContained(false);
            }
          }
        }
        approved = true;
        setVisible(false);
      }
    });
    button = new JButton(LaunchRes.getString("Saver.Button.Cancel"));
    buttonbar.add(button);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        editor.getCellEditorValue(); // make sure edits are saved
        revert();
        setVisible(false);
      }
    });
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        if (!isApproved()) revert();
      }
    });
  }

  /**
   * Creates a new tree.
   */
  private void createTree() {
    // create root node
    root = new SaverNode();
    // add child nodes to root
    int n = launcher.tabbedPane.getTabCount();
    for (int i = 0; i < n; i++) {
      SaverNode tab = new SaverNode(launcher.getTab(i).getRootNode());
      root.add(tab);
      // add tab's owned nodes
      LaunchNode[] nodes = tab.node.getOwnedNodes();
      for (int k = 0; k < nodes.length; k++) {
        SaverNode child = new SaverNode(nodes[k]);
        if (child.node != tab.node)
          tab.add(child);
      }
    }
    // create tree model and tree
    TreeModel treeModel = new DefaultTreeModel(root);
    tree = new JTree(treeModel);
    tree.setCellRenderer(new SaveRenderer());
//    tree.setCellRenderer(new SaverRenderer());
//    editor.node = null;
//    tree.setCellEditor(editor);
//    tree.setEditable(true);
    tree.setShowsRootHandles(true);
    tree.getSelectionModel().setSelectionMode
        (TreeSelectionModel.SINGLE_TREE_SELECTION);
    tree.addTreeSelectionListener(new TreeSelectionListener() {
      public void valueChanged(TreeSelectionEvent e) {
        refresh();
      }
    });
    // put tree in a scroller and add to dialog
    if (treeScroller != null) getContentPane().remove(treeScroller);
    treeScroller = new JScrollPane(tree);
    getContentPane().add(treeScroller, BorderLayout.CENTER);
    // expand tree and select root node
    tree.expandRow(tree.getRowCount() - 1);
    setSelectedNode(root);
  }

  /**
   * Selects a path with a chooser.
   */
  private void choosePath() {
    JFileChooser chooser = Launcher.getXMLChooser();
    if (getSelectedNode() == root) {
      chooser.setSelectedFile(new File(launcher.tabSetName));
      chooser.setFileFilter(Launcher.xsetFileFilter);
      int result = chooser.showDialog(this, LaunchRes.getString(
          "Saver.FileChooser.Title"));
      if (result == JFileChooser.APPROVE_OPTION) {
        File file = chooser.getSelectedFile();
        // force "xset" extension
        if (!"xset".equals(Launcher.getExtension(file.getName()))) {
          String fileName = file.getPath();
          while (fileName.indexOf(".") > -1) {
            fileName = fileName.substring(0, fileName.length() - 1);
          }
          fileName += ".xset";
          file = new File(fileName);
        }
        String path = Launcher.forwardSlash(file.getAbsolutePath());
        launcher.tabSetBasePath = XML.getDirectoryPath(path);
        launcher.tabSetName = Launcher.getName(path);
        pathField.setText(path);
      }
    }
    refresh();
  }

  /**
   * A tree node for the saver tree.
   */
  private class SaverNode extends DefaultMutableTreeNode {
    private LaunchNode node;
    boolean save = true;
    SaverNode() {}
    SaverNode(LaunchNode node) {
      this.node = node;
      save = node.saving;
    }
    public String toString() {
      if (node == null)
        return LaunchRes.getString("Saver.Tree.TabSet") +
          "  " + launcher.tabSetName;
      else if (node.isRoot()) {
        return LaunchRes.getString("Saver.Tree.Tab") +
            "  " + node.getFileName();
      }
      else {
        return LaunchRes.getString("Saver.Tree.Node") +
            "  " + node.getFileName();
      }
    }
  }

  /**
   * A cell editor to edit tree nodes.
   */
  private class SaverEditor
      extends AbstractCellEditor
      implements TreeCellEditor {

    SaverRenderer renderer = new SaverRenderer();
    SaverNode node;

    public Component getTreeCellEditorComponent(JTree tree, Object value,
                                         boolean isSelected, boolean expanded,
                                         boolean leaf, int row) {
      Component cell = renderer.getTreeCellRendererComponent(tree, value,
                                     isSelected, expanded, leaf, row, false);
      node = (SaverNode) value;
      return cell;
    }

    public Object getCellEditorValue() {
      if (node != null) {
        node.save = renderer.checkBox.isSelected();
        if (node.node != null) node.node.saving = node.save;
        node = null;
      }
      return null;
    }
  }

  /**
   * A cell renderer to show tree nodes.
   */
  private class SaveRenderer extends DefaultTreeCellRenderer {

    public Component getTreeCellRendererComponent(
        JTree tree, Object value, boolean sel, boolean expanded,
        boolean leaf, int row, boolean hasFocus) {

      super.getTreeCellRendererComponent(
          tree, value, sel, expanded, leaf, row, hasFocus);
      SaverNode node = (SaverNode) value;
      File file;
      if (node == root) {
        String path = XML.getResolvedPath(launcher.tabSetName, Launcher.tabSetBasePath);
        Resource res = ResourceLoader.getResource(path);
        file = res == null? null: res.getFile();
      }
      else {
        String path = XML.getResolvedPath(node.node.getFileName(), Launcher.tabSetBasePath);
        file = node.node.getFile();
        setToolTipText(path);
      }
      // green icon if new file will be written
      if (file == null) {
        setIcon(node != root? Launcher.greenFileIcon: Launcher.greenFolderIcon);
      }
      // yellow icon if existing writable file will be overwritten
      else if (file.canWrite()) {
        setIcon(node != root? Launcher.yellowFileIcon: Launcher.yellowFolderIcon);
      }
      // red icon if existing read-only file exists
      else {
        setIcon(node != root? Launcher.redFileIcon: Launcher.redFolderIcon);
      }
      return this;
    }
  }

  /**
   * A renderer with checkBoxes.
   */
  class SaverRenderer implements TreeCellRenderer {
    JPanel panel = new JPanel(new BorderLayout());
    SaveRenderer renderer = new SaveRenderer();
    JCheckBox checkBox = new JCheckBox();

    public SaverRenderer() {
      checkBox.setOpaque(false);
      panel.add(checkBox, BorderLayout.WEST);
      panel.setOpaque(false);
    }

    public Component getTreeCellRendererComponent(JTree tree, Object value,
        boolean selected, boolean expanded, boolean leaf, int row,
        boolean hasFocus) {
      Component cell = renderer.getTreeCellRendererComponent(
          tree, value, selected, expanded, leaf, row, hasFocus);
      SaverNode node = (SaverNode)value;
      checkBox.setSelected(node.save);
      panel.add(cell, BorderLayout.CENTER);
      panel.setEnabled(tree.isEnabled());
      return panel;
    }
  }
}
